from masced_bandits.bandits.egreedy import egreedy
from masced_bandits.bandits.EXP3 import EXP3
from masced_bandits.bandits.EXP4 import EXP4
from masced_bandits.bandits.UCB import UCB
from masced_bandits.bandits.UCBImproved import UCBImproved
from masced_bandits.bandits.SWUCB import SWUCB
from masced_bandits.bandits.EwS import EwS
from masced_bandits.bandits.EXP3S import EXP3S
from masced_bandits.bandits.DUCB import DUCB
from masced_bandits.bandits.ETC import ETC

import numpy as np


def init_bandit(**kwargs):
    chosen_bandit = {
        "egreedy": egreedy,
        "UCB": UCB,
        "EXP3": EXP3,
        "UCBImproved": UCBImproved,
        "SWUCB": SWUCB,
        "EwS": EwS,
        "EXP3S": EXP3S,
        "EXP4": EXP4,
        "DUCB": DUCB,
        "ETC": ETC

    }.get(kwargs["name"], None)

    if(chosen_bandit):
        return chosen_bandit(**kwargs)
    else:
        raise RuntimeError("Specified Bandit " + str(chosen_bandit) + " did not exist")



from masced_bandits.bandits.Bandit import Bandit
import numpy as np

CUM_REWARD = 0
N_K = 1

class egreedy(Bandit):
    def __init__(self, **kwargs):
        super().__init__("e-greedy")

        self.game_list = []
        self.epsilon = float(kwargs.get('epsilon',0.5))
        self.decay_rate = float(kwargs.get('decay_rate',1.0))

        self.arm_reward_pairs = {}
        for arm in self.arms: self.arm_reward_pairs[arm] = [0.0,0.0]

    def get_next_arm(self, reward):
        self.arm_reward_pairs[self.last_action][CUM_REWARD]+=reward
        self.arm_reward_pairs[self.last_action][N_K]+=1

        choice = np.random.random()

        if(choice < self.epsilon):    
            next_arm = self.arms[np.random.choice(len(self.arms))]
        else:
            next_arm = max(self.arms, key=lambda arm: self.reward_average(arm))

        self.epsilon = self.epsilon/self.decay_rate

        self.last_action = next_arm
        return next_arm

    def reward_average(self, arm):
        return self.arm_reward_pairs[arm][CUM_REWARD] / self.arm_reward_pairs[arm][N_K] if self.arm_reward_pairs[arm][N_K] else 0


# Abstract interface for a execution strategy
#
# An execution strategy that forms the logic of the experiment.
from masced_bandits.bandit_options import bandit_args
import numpy as np

class Bandit:
    def __init__(self, name):
        self.name = name
        self.last_action = bandit_args["initial_configuration"]

        if(bandit_args['arms']):
            self.arms = bandit_args['arms']
        else:
            print("no arms specified, in Bandit constructor")
            raise RuntimeError("No arms specified")
    def get_next_arm(self, reward):
        """ starts execution """
        pass
    
    def visualize(self):
        """ will create a visualization of the current state of the algorithm """
        pass







# Abstract interface for an expert to be used in EXP4
#
# Provides a distribution over some arms which can be updated and drawn from.
import numpy as np

class Expert:
    def __init__(self):
        self.num_arms = None
        self.distribution = None #should be initialized in subclass
        self.weights = None #should be initialized in subclass

    def set_functions(self, formula):
        'implemented in subclass'
        pass

    def sample_action(self):
        
        choice = np.random.choice(np.arange(0, self.num_arms), p= self.distribution)
        return choice

    def propagate_reward(self, reward, chosen_action):
        'should be implemented in the subclass'
        pass
    






import numpy as np
from masced_bandits.bandits.Bandit import Bandit
from collections import deque


REWARD = 0
ACTION = 1
XI = 2


class SWUCB(Bandit):
    def __init__(self, **kwargs):
        super().__init__("SWUCB-" + str(kwargs))

        self.look_back = int(kwargs.get("look_back",10))

        self.bandit_round = -1
        self.game_list = deque(maxlen=self.look_back)

    def get_next_arm(self, reward):
               
        self.game_list.append([reward, self.last_action])

        self.bandit_round = self.bandit_round + 1

        if((self.bandit_round) < (len(self.arms))):
            #initial exploration      
            next_arm = self.arms[self.bandit_round]
        else:
            next_arm = max(self.arms, key=lambda arm: \
                       self.X_t(arm) + self.c_t(arm))

        self.last_action = next_arm
        return next_arm

    def N_t(self, arm):
        count = 0
        for game in self.game_list:
            if(game[ACTION] == arm): count += 1

        return count

    def X_t(self, arm):

        summated = 0

        
        for current_game in self.game_list:
            if(current_game[ACTION] == arm): #the games in which i was the arm
                X_s = current_game[REWARD]
                summated+=X_s
        try:
            times_arm_played = self.N_t(arm) 
            if(summated == 0 or times_arm_played == 0): return 0
            
            return summated/self.N_t(arm)
        except:
            print("Divide by zero error likely")
            print(self.game_list)
            print("Last tau games are")
            print(self.game_list[-self.look_back:])
            print("Length of game list is ")
            print(len(self.game_list))
            print("Result of N_t that was trigger is ")
            print(self.N_t(arm))
            print("and the arm causing it was ")
            print(arm)
            print("Value of summated is ")
            print(summated)
            exit(1)

        
    def c_t(self, arm):
        res = self.chapter7(arm, len(self.game_list)) #t or tau has become unnecessary

        return res
    
    def sqreward_average(self, arm):
        r_sum = 0

        for game in self.game_list:  
            if(game[ACTION] == arm):
                r_sum+=np.square(game[REWARD])

        times_arm_played = self.N_t(arm) 
        if(r_sum == 0 or times_arm_played == 0): return 0
            
        return r_sum/times_arm_played
    
    def tuned(self, arm, n):
        n_k = self.N_t(arm)

        average_of_squares = self.sqreward_average(arm)
        square_of_average = np.square(self.X_t(arm))
        estimated_variance = average_of_squares - square_of_average
        param = np.log(n) / n_k
        V_k = estimated_variance + np.sqrt(2 * param)

        return np.sqrt(param * V_k)
    
    def chapter7(self, arm, n):
        DELTA = 1 / np.square(self.look_back)

        n_k = self.N_t(arm)


        upper_term = 2 * np.log( (1 / DELTA) )
        
        to_be_sqrt = upper_term/n_k
        
        return np.sqrt(to_be_sqrt)
        





from masced_bandits.bandits.Bandit import Bandit
import numpy as np


FORMULA_FUNC = None

CUM_REWARD = 0
CUM_SQ_REWARD = 1
N_K = 2




trace_len = 15000 #the total time of chosen trace in SWIM in seconds
total_count = round(trace_len / 60) 
DELTA = 1 / np.square(total_count)


class UCB(Bandit):
    def __init__(self, **kwargs):
        formula = kwargs.get("formula","TN")
        super().__init__("UCB-" + formula)
        self.formula = self.formula_to_function(formula)
        
        self.bandit_round = -1
        self.arm_reward_pairs = {}
        for arm in self.arms: self.arm_reward_pairs[arm] = [0.0,0.0,0.0]
        


        
    def get_next_arm(self, reward):
        self.arm_reward_pairs[self.last_action][CUM_REWARD]+=reward
        self.arm_reward_pairs[self.last_action][CUM_SQ_REWARD]+=np.square(reward)
        self.arm_reward_pairs[self.last_action][N_K]+=1

        self.bandit_round+=1

        if((self.bandit_round) < (len(self.arms))):    
            next_arm = self.arms[self.bandit_round]
        else:
            next_arm = max(self.arms, key=lambda arm: \
                       self.reward_average(arm) + self.formula(arm, self.bandit_round))

        self.last_action = next_arm
        return next_arm


    def formula_to_function(self, choice):
        funcs = {
                "FH": self.chapter7,
                "AO": self.asymptotically_optimal,
                "OG": self.auer2002UCB,
                "TN": self.tuned
            }
            
        func = funcs.get(choice)
        #print(func.__doc__)
        return func

    def reward_average(self, arm):
        return self.arm_reward_pairs[arm][CUM_REWARD] / self.arm_reward_pairs[arm][N_K]


    def chapter7(self, arm, n):
        global DELTA

        upper_term = 2 * np.log( (1 / DELTA) )
        
        to_be_sqrt = upper_term/self.arm_reward_pairs[arm][N_K]
        
        return np.sqrt(to_be_sqrt)

    def asymptotically_optimal(self, arm, t):
        T_i = self.arm_reward_pairs[arm][N_K]

        f_t = 1 + (t  * np.square(np.log(t)))

        upper_term = 2 * (np.log(f_t))

        lower_term =  T_i

        to_be_sqrt = upper_term/lower_term
        
        return np.sqrt(to_be_sqrt)

    def auer2002UCB(self, arm, t):
        T_i = self.arm_reward_pairs[arm][N_K]

        upper_term = 2 * (np.log(t))

        lower_term =  T_i

        to_be_sqrt = upper_term/lower_term
        
        return np.sqrt(to_be_sqrt)

    def tuned(self, arm, n):
        n_k = self.arm_reward_pairs[arm][N_K]

        average_of_squares = self.arm_reward_pairs[arm][CUM_SQ_REWARD] / n_k
        square_of_average = np.square(self.arm_reward_pairs[arm][CUM_REWARD]/n_k)
        estimated_variance = average_of_squares - square_of_average
        param = np.log(n) / n_k
        V_k = estimated_variance + np.sqrt(2 * param)

        confidence_value = np.sqrt(param * V_k)
        if(np.isnan(confidence_value)):
            print("\n\n\n\ncaught a nan\n\n\n\n\n")
            return 0
        else:
            return confidence_value




import numpy as np
from masced_bandits.bandits.Bandit import Bandit
from masced_bandits.bandits.Expert import Expert

CUM_REWARD = 0
N_K = 1

class EwS(Bandit, Expert):
    def __init__(self, **kwargs): 
        super().__init__("EwS")
        self.weights, self.distribution = self.ews_initialize(len(self.arms))
        self.num_arms = len(self.arms)

        self.arm_reward_pairs = {}
        for arm in self.arms: self.arm_reward_pairs[arm] = [0.0,0.0]

        self.distr_func()

           
    def ews_initialize(self, num_arms):
        return [0] * num_arms, []


    def get_next_arm(self, reward):

        self.arm_reward_pairs[self.last_action][CUM_REWARD]+=reward
        self.arm_reward_pairs[self.last_action][N_K]+=1

        self.update_func() #Update weights

        self.distr_func() #(re-)calculate Pt

        new_action = self.sample_action()
  
        self.last_action = self.arms[new_action]

        return self.arms[new_action]

    def distr_func(self):
        # exp(eta * S^_t-1i) / SUMkj=1 exp(eta * S^_t-1j)
        sum_weights = 0.0

        for i, arm in enumerate(self.arms):
            times_arm_played = self.arm_reward_pairs[arm][N_K]
            squared_gap = np.square(self.weights[i])
            sum_weights+= np.exp(-2*times_arm_played*squared_gap)

        self.distribution.clear()

        for i, arm in enumerate(self.arms):
            times_arm_played = self.arm_reward_pairs[arm][N_K]
            squared_gap = np.square(self.weights[i])
            self.distribution.append(np.exp(-2 * times_arm_played * squared_gap)/sum_weights)

    def update_func(self, payoff=None, action=None):
        empirical_gaps = None

        empirical_mean_rewards = []
        for arm in self.arms:
            current_entry = self.arm_reward_pairs[arm]
            if(current_entry[N_K] == 0):
                empirical_mean_rewards.append(0)
            else:
                empirical_mean_rewards.append(current_entry[CUM_REWARD]/current_entry[N_K])

        empirical_gaps = [max(empirical_mean_rewards) - emp_mean_rew for emp_mean_rew in empirical_mean_rewards]

        self.weights = empirical_gaps

    def propagate_reward(self, reward, chosen_action):
        self.arm_reward_pairs[chosen_action][CUM_REWARD]+=reward
        self.arm_reward_pairs[chosen_action][N_K]+=1

        self.update_func()

        self.distr_func()
